"""DLNA service."""
from ..service import Service
from .browser import DlnaBrowser


class DlnaService(Service):
    """
    Dlna service.
    """

    def __init__(self, api):
        super().__init__(DlnaBrowser(api))

    @property
    def name(self):
        """
        :returns: Name of the service
        """
        return 'dlna'



"""Speaker group."""
from .base import SpeakerBase
from .clock import ClockGroup
from .equalizer import EqualizerGroup


class SpeakerGroup(SpeakerBase):
    """
    Speaker group.

    Use Speaker.group() to initiate grouping.
    """

    def __init__(self, api, name, speakers):
        """
        The first speaker should be the main speaker controlling the playback.

        :param api: Api to control the main speaker
        :param name: Name to used for the group
        :param speaker: List of Speaker instances in this group
        """
        self._api = api
        self._name = name
        self._speakers = speakers

    @property
    def ip_address(self):
        """
        :returns: Speaker's ip address
        """
        return self._speakers[0].ip_address

    @property
    def mac_address(self):
        """
        :returns: Speaker's mac address
        """
        return self._speakers[0].mac_address

    @property
    def speakers(self):
        """
        :returns: List of speakers in this group.
        """
        return self._speakers

    def get_name(self):
        """
        :returns: Group's name
        """
        return self._name

    def set_name(self, name):
        """
        Set group name.

        :param name: New group name
        """
        self._api.set_group_name(name)
        self._name = name

    def get_volume(self):
        """
        Get current volume.

        It uses main speaker volume as a reference.

        :returns: int current volume
        """
        return self._speakers[0].get_volume()

    def set_volume(self, volume):
        """
        Set volume for all speakers in the group.

        It uses main speaker's volume as a reference point and updates secondary speakers proportionally.

        :param volume: speaker volume, integer between 0 and 100
        """
        ratio = volume / self._speakers[0].get_volume()

        self._speakers[0].set_volume(volume)

        for speaker in self._speakers[1:]:
            speaker.set_volume(min(100, int(speaker.get_volume() * ratio)))

    def get_sources(self):
        """
        Get all supported sources.

        :returns: List of supported sources.
        """
        return self._speakers[0].get_sources()

    def get_source(self):
        """
        Get currently selected source.

        :returns: selected source string.
        """
        return self._speakers[0].get_source()

    def set_source(self, source):
        """
        Set speaker source.

        :param source: Speaker source, one of returned by get_sources()
        """
        return self._speakers[0].set_source(source)

    def is_muted(self):
        """
        Check if all speakers in this group are muted.

        :returns: True if muted, False otherwise
        """
        for speaker in self._speakers:
            if not speaker.is_muted():
                return False

        return True

    def mute(self):
        """Mute all speakers in this group."""
        for speaker in self._speakers:
            speaker.mute()

    def unmute(self):
        """Unmute all speakers in this group."""
        for speaker in self._speakers:
            speaker.unmute()

    def get_services_names(self):
        """
        Get all supported services names.

        :returns: List of strings
        """
        return self._speakers[0].get_services_names()

    def service(self, name):
        """
        Get service by type

        :returns: Service instance
        """
        return self._speakers[0].service(name)

    @property
    def clock(self):
        """
        Get clock to control time functions.

        :returns: Clock instance
        """
        return ClockGroup([s.clock for s in self._speakers])

    @property
    def equalizer(self):
        """
        Get equalizer to control sound adjustments.

        :returns: Equalizer instance
        """
        return EqualizerGroup([s.equalizer for s in self._speakers])

    @property
    def player(self):
        """
        Get currently active player.

        Use player to control playback e.g. pause, resume, get track info etc.

        :returns: Player instance
        """
        return self._speakers[0].player

    def browser(self, name):
        """
        Get media browser by type

        :returns: Browser instance
        """
        return self._speakers[0].browser(name)

    def group(self, name, speakers):
        """
        Adds speakers to this group.

        This group's main speaker will continue to be the main speaker controlling the playback.

        :param speaker: List of Speaker instances to add to this group
        :returns: This speaker group instance
        """
        if isinstance(speakers, SpeakerBase):
            speakers = [speakers]

        main_speaker = self._speakers[0]
        speakers = self._speakers[1:] + speakers

        speaker_group = main_speaker.group(name, speakers)

        self._name = speaker_group.get_name()
        self._speakers = speaker_group.speakers

        return self

    def ungroup(self):
        """
        Remove all speakers from this group and disband it.
        """
        for speaker in reversed(self._speakers):
            speaker.ungroup()



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import REPEAT_ALL
from samsung_multiroom.service import REPEAT_ONE
from samsung_multiroom.service.dlna import DlnaPlayer


def _get_player():
    api = MagicMock()

    player = DlnaPlayer(api)

    return (player, api)


class TestDlnaPlayer(unittest.TestCase):

    def test_play(self):
        playlist = [
            type('Item', (object, ), {
                'object_id': 'id1',
                'object_type': 'some_type',
                'title': 'title 1',
            }),
            type('Item', (object, ), {
                'device_udn': 'device_udn',
                'object_id': 'id2',
                'object_type': 'dlna_audio',
                'title': 'track 2',
                'artist': 'artist 2',
                'thumbnail_url': 'thumb 2'
            }),
            type('Item', (object, ), {
                'device_udn': 'device_udn',
                'object_id': 'id3',
                'object_type': 'dlna_audio',
                'title': 'track 3',
                'artist': 'artist 3',
                'thumbnail_url': 'thumb 3'
            }),
            type('Item', (object, ), {
                'object_id': 'id4',
                'object_type': 'some_type2',
                'title': 'title 4',
            })
        ]

        player, api = _get_player()

        self.assertTrue(player.play(playlist))

        items = [
            {
                'device_udn': 'device_udn',
                'object_id': 'id2',
                'title': 'track 2',
                'artist': 'artist 2',
                'thumbnail': 'thumb 2'
            },
            {
                'device_udn': 'device_udn',
                'object_id': 'id3',
                'title': 'track 3',
                'artist': 'artist 3',
                'thumbnail': 'thumb 3'
            }
        ]

        api.set_playlist_playback_control.assert_called_once_with(items)

    def test_play_returns_false_for_unsupported_playlist(self):
        playlist = [
            type('Item', (object, ), {
                'object_id': 'id1',
                'object_type': 'some_type',
                'title': 'title 1',
            }),
            type('Item', (object, ), {
                'object_id': 'id4',
                'object_type': 'some_type2',
                'title': 'title 4',
            })
        ]

        player, api = _get_player()

        self.assertFalse(player.play(playlist))

        api.set_playlist_playback_control.assert_not_called()

    def test_jump(self):
        player, api = _get_player()

        player.jump(50)

        api.set_search_time.assert_called_once_with(50)

    def test_resume(self):
        player, api = _get_player()

        player.resume()

        api.set_playback_control.assert_called_once_with('resume')

    @unittest.skip('Pending implementation')
    def test_stop(self):
        player, api = _get_player()

        player.stop()

    def test_pause(self):
        player, api = _get_player()

        player.pause()

        api.set_playback_control.assert_called_once_with('pause')

    def test_next(self):
        player, api = _get_player()

        player.next()

        api.set_trick_mode.assert_called_once_with('next')

    def test_previous(self):
        player, api = _get_player()

        player.previous()

        api.set_trick_mode.assert_called_once_with('previous')

    def test_repeat(self):
        player, api = _get_player()

        player.repeat(REPEAT_ALL)

        api.set_repeat_mode.assert_called_once_with('all')

    def test_repeat_raises_value_error(self):
        player, api = _get_player()

        self.assertRaises(ValueError, player.repeat, 'unsupported_mode')

        api.set_repeat_mode.assert_not_called()

    def test_get_repeat(self):
        player, api = _get_player()
        api.get_repeat_mode.return_value = 'one'

        repeat = player.get_repeat()

        self.assertEqual(repeat, REPEAT_ONE)

        api.get_repeat_mode.assert_called_once()

    def test_get_current_track(self):
        player, api = _get_player()
        api.get_music_info.return_value = {
            'device_udn': 'uuid:00113249-398f-0011-8f39-8f3949321100',
            'playertype': 'allshare',
            'playbacktype': 'folder',
            'sourcename': None,
            'parentid': '22$30224',
            'parentid2': None,
            'playindex': '8',
            'objectid': '22$@52947',
            'title': 'New star in the sky',
            'artist': 'Air',
            'album': 'Moon Safari',
            'thumbnail': 'http://192.168.1.111:50002/transcoder/jpegtnscaler.cgi/folderart/52947.jpg',
            'timelength': '0:05:40.000',
            'playtime': '325067',
            'seek': 'enable',
            'pause': 'enable',
        }

        track = player.get_current_track()

        api.get_music_info.assert_called_once()

        self.assertEqual(track.title, 'New star in the sky')
        self.assertEqual(track.artist, 'Air')
        self.assertEqual(track.album, 'Moon Safari')
        self.assertEqual(track.duration, 340)
        self.assertEqual(track.position, 325)
        self.assertEqual(track.thumbnail_url, 'http://192.168.1.111:50002/transcoder/jpegtnscaler.cgi/folderart/52947.jpg')
        self.assertEqual(track.device_udn, 'uuid:00113249-398f-0011-8f39-8f3949321100')
        self.assertEqual(track.object_id, '22$@52947')
        self.assertEqual(track.object_type, 'dlna_audio')

    def test_is_supported(self):
        player, api = _get_player()

        self.assertTrue(player.is_supported('wifi', 'dlna'))
        self.assertFalse(player.is_supported('wifi', 'cp'))
        self.assertFalse(player.is_supported('bt'))



"""
Parse API XML response.
"""
import xmltodict


class ApiResponse:
    """
    Extract key information from api response body text.
    """

    def __init__(self, response_text):
        self._name = None
        self._success = None
        self._data = None
        self._raw = None

        self._parse(response_text)

    @property
    def name(self):
        """
        :returns: resopnse type/name
        """
        return self._name

    @property
    def success(self):
        """
        :returns: True if response was successful
        """
        return self._success

    @property
    def data(self):
        """
        :returns: response data
        """
        return self._data

    @property
    def raw(self):
        """
        :returns: Raw response text
        """
        return self._raw

    def _parse(self, response_text):
        self._success = False
        self._raw = response_text

        try:
            response_dict = xmltodict.parse(response_text)
        except xmltodict.expat.ExpatError:
            return

        # for some requests speaker returns command in response that does not match request command
        try:
            response_command = next(iter(response_dict))

            self._name = response_dict[response_command]['method']
            self._data = response_dict[response_command]['response']
            self._success = (self._data['@result'] == 'ok')

            # redundant
            del self._data['@result']

            return
        except KeyError:
            pass



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import ServiceRegistry
from samsung_multiroom.service.app import AppService
from samsung_multiroom.service.dlna import DlnaService
from samsung_multiroom.service.tunein import TuneInService


def _get_service_registry():
    api = MagicMock()
    api.get_cp_list.return_value = [
        {
            'cpid': '0',
            'cpname': 'Signed out service',
            'signinstatus': '0',
        },
        {
            'cpid': '1',
            'cpname': 'Signed in service',
            'signinstatus': '1',
            'username': 'service_username',
        },
        {
            'cpid': '2',
            'cpname': 'Service 2',
            'signinstatus': '0',
        },
        {
            'cpid': '3',
            'cpname': 'Service 3',
            'signinstatus': '0',
        }
    ]

    service_registry = ServiceRegistry(api)

    return (service_registry, api)


class TestServiceRegistry(unittest.TestCase):

    @unittest.mock.patch('inspect.signature')
    def test_get_services_names(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        service_registry, api = _get_service_registry()

        services_names = service_registry.get_services_names()

        self.assertEqual(services_names, ['dlna', 'tunein', 'Signed out service', 'Signed in service', 'Service 2', 'Service 3'])

    @unittest.mock.patch('inspect.signature')
    def test_service_dlna(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        service_registry, api = _get_service_registry()

        service = service_registry.service('dlna')

        self.assertIsInstance(service, DlnaService)
        self.assertEqual(service.name, 'dlna')

    @unittest.mock.patch('inspect.signature')
    def test_service_tunein(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        service_registry, api = _get_service_registry()

        service = service_registry.service('tunein')

        self.assertIsInstance(service, TuneInService)
        self.assertEqual(service.name, 'tunein')

    @unittest.mock.patch('inspect.signature')
    def test_service_app(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        service_registry, api = _get_service_registry()

        service = service_registry.service('Service 2')

        self.assertIsInstance(service, AppService)
        self.assertEqual(service.name, 'Service 2')



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.clock import ClockGroup
from samsung_multiroom.equalizer import EqualizerGroup
from samsung_multiroom.group import SpeakerGroup


def _get_speaker_group():
    api = MagicMock()

    speakers = [
        MagicMock(),
        MagicMock(),
        MagicMock(),
    ]

    name = 'Test group'

    speaker_group = SpeakerGroup(api, name, speakers)

    return (speaker_group, api, speakers)


class TestSpeakerGroup(unittest.TestCase):

    def test_ip_address(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].ip_address = '192.168.1.129'

        ip_address = speaker_group.ip_address

        self.assertEqual(ip_address, '192.168.1.129')

    def test_mac_address(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].mac_address = '00:11:22:33:44:55'

        mac_address = speaker_group.mac_address

        self.assertEqual(mac_address, '00:11:22:33:44:55')

    def test_get_name(self):
        speaker_group, api, speakers = _get_speaker_group()

        name = speaker_group.get_name()

        self.assertEqual(name, 'Test group')

    def test_set_name(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.set_name('Updated group name')
        name = speaker_group.get_name()

        self.assertEqual(name, 'Updated group name')

        api.set_group_name.assert_called_once_with('Updated group name')

    def test_get_volume_returns_main_speaker_volume(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].get_volume.return_value = 13

        volume = speaker_group.get_volume()

        self.assertEqual(volume, 13)
        speakers[0].get_volume.assert_called_once()
        speakers[1].get_volume.assert_not_called()
        speakers[2].get_volume.assert_not_called()

    def test_set_volume_updates_proportionally(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].get_volume.return_value = 20
        speakers[1].get_volume.return_value = 10
        speakers[2].get_volume.return_value = 5

        speaker_group.set_volume(40)

        speakers[0].set_volume.assert_called_once_with(40)
        speakers[1].set_volume.assert_called_once_with(20)
        speakers[2].set_volume.assert_called_once_with(10)

    def test_set_volume_stays_within_bounds(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].get_volume.return_value = 20
        speakers[1].get_volume.return_value = 60
        speakers[2].get_volume.return_value = 60

        speaker_group.set_volume(40)

        speakers[0].set_volume.assert_called_once_with(40)
        speakers[1].set_volume.assert_called_once_with(100)
        speakers[2].set_volume.assert_called_once_with(100)

    def test_get_sources(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].get_sources.return_value = ['aux', 'bt', 'hdmi', 'optical', 'soundshare', 'wifi']

        sources = speaker_group.get_sources()

        self.assertEqual(sorted(sources), sorted(['aux', 'bt', 'hdmi', 'optical', 'soundshare', 'wifi']))

        speakers[0].get_sources.assert_called_once()
        speakers[1].get_sources.assert_not_called()
        speakers[2].get_sources.assert_not_called()

    def test_get_source(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].get_source.return_value = 'wifi'

        source = speaker_group.get_source()

        self.assertEqual(source, 'wifi')

        speakers[0].get_source.assert_called_once()
        speakers[1].get_source.assert_not_called()
        speakers[2].get_source.assert_not_called()

    def test_set_source(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.set_source('hdmi')

        speakers[0].set_source.assert_called_once_with('hdmi')
        speakers[1].set_source.assert_not_called()
        speakers[2].set_source.assert_not_called()

    def test_is_muted_returns_true_if_all_muted(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].is_muted.return_value = True
        speakers[1].is_muted.return_value = True
        speakers[2].is_muted.return_value = True

        muted = speaker_group.is_muted()

        self.assertTrue(muted)

        speakers[0].is_muted.assert_called_once()
        speakers[1].is_muted.assert_called_once()
        speakers[2].is_muted.assert_called_once()

    def test_is_muted_returns_false_if_one_unmuted(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].is_muted.return_value = True
        speakers[1].is_muted.return_value = False
        speakers[2].is_muted.return_value = True

        muted = speaker_group.is_muted()

        self.assertFalse(muted)

        speakers[0].is_muted.assert_called_once()
        speakers[1].is_muted.assert_called_once()
        speakers[2].is_muted.assert_not_called()

    def test_mute_mutes_all_speakers(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.mute()

        speakers[0].mute.assert_called_once()
        speakers[1].mute.assert_called_once()
        speakers[2].mute.assert_called_once()

    def test_unmute_unmutes_all_speakers(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.unmute()

        speakers[0].unmute.assert_called_once()
        speakers[1].unmute.assert_called_once()
        speakers[2].unmute.assert_called_once()

    def test_get_services_names_returns_main_speaker_services(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.get_services_names()

        speakers[0].get_services_names.assert_called_once()
        speakers[1].get_services_names.assert_not_called()
        speakers[2].get_services_names.assert_not_called()

    def test_service_returns_main_speaker_service(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.service('dlna')

        speakers[0].service.assert_called_once_with('dlna')
        speakers[1].service.assert_not_called()
        speakers[2].service.assert_not_called()

    def test_player_returns_main_speaker_player(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].player = MagicMock()

        player = speaker_group.player

        self.assertEqual(player, speakers[0].player)

    def test_browser_returns_main_speaker_browser(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].browser.return_value = MagicMock()

        browser = speaker_group.browser('b2')

        self.assertEqual(browser, speakers[0].browser.return_value)

        speakers[0].browser.assert_called_once_with('b2')

    def test_equalizer(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].equalizer = MagicMock()
        speakers[1].equalizer = MagicMock()
        speakers[2].equalizer = MagicMock()

        equalizer = speaker_group.equalizer

        self.assertIsInstance(equalizer, EqualizerGroup)
        self.assertEqual(equalizer.equalizers, [speakers[0].equalizer, speakers[1].equalizer, speakers[2].equalizer])

    def test_clock(self):
        speaker_group, api, speakers = _get_speaker_group()
        speakers[0].clock = MagicMock()
        speakers[1].clock = MagicMock()
        speakers[2].clock = MagicMock()

        clock = speaker_group.clock

        self.assertIsInstance(clock, ClockGroup)
        self.assertEqual(clock.clocks, [speakers[0].clock, speakers[1].clock, speakers[2].clock])

    def test_group_adds_speakers_to_existing_group(self):
        speaker_group, api, speakers = _get_speaker_group()

        new_speakers = [
            MagicMock(),
            MagicMock(),
        ]

        speakers[0].group.return_value = MagicMock()
        speakers[0].group.return_value.get_name.return_value = 'Expanded group'
        speakers[0].group.return_value.speakers = speakers + new_speakers

        new_speaker_group = speaker_group.group('Expanded group', new_speakers)

        self.assertEqual(new_speaker_group, speaker_group)
        self.assertEqual(new_speaker_group.get_name(), 'Expanded group')
        self.assertEqual(new_speaker_group.speakers, speakers + new_speakers)

        speakers[0].group.assert_called_once_with('Expanded group', speakers[1:] + new_speakers)

    def test_ungroup_ungroups_all_speakers(self):
        speaker_group, api, speakers = _get_speaker_group()

        speaker_group.ungroup()

        speakers[0].ungroup.assert_called_once()
        speakers[1].ungroup.assert_called_once()
        speakers[2].ungroup.assert_called_once()



import datetime
import unittest
from unittest.mock import MagicMock

from samsung_multiroom.clock import Alarm
from samsung_multiroom.clock import AlarmSlot


def get_alarm():
    api = MagicMock()
    api.get_alarm_info.return_value = [
        {
            '@index': '0',
            'hour': '10',
            'min': '15',
            'week': '0x40',
            'volume': '20',
            'title': None,
            'description': None,
            'thumbnail': None,
            'stationurl': None,
            'set': 'on',
            'soundenable': 'on',
            'sound': '1',
            'alarmsoundname': 'Disco',
            'duration': '10',
        },
        {
            '@index': '1',
            'hour': '8',
            'min': '5',
            'week': '0x3e',
            'volume': '5',
            'title': 'BBC Radio 4',
            'description': 'Intelligent speech',
            'thumbnail': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
            'stationurl': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
            'set': 'on',
            'soundenable': 'off',
            'sound': '-1',
            'alarmsoundname': '',
            'duration': '0',
        }
    ]

    alarm = Alarm(api)

    return (alarm, api)


NOW = datetime.datetime(2018, 1, 7, 15, 45, 32)


class TestAlarm(unittest.TestCase):

    def test_get_slots(self):
        alarm, api = get_alarm()

        slots = alarm.get_slots()

        self.assertEqual(len(slots), 3)

        # 0
        self.assertEqual(slots[0].index, 0)
        self.assertEqual(slots[0].time, '10:15')
        self.assertEqual(slots[0].weekdays, [6]) # Sunday
        self.assertEqual(slots[0].volume, 20)
        self.assertEqual(slots[0].duration, 10)
        self.assertEqual(slots[0].station_data, {
            'title': None,
            'description': None,
            'thumbnail_url': None,
            'station_url': None,
        })
        self.assertTrue(slots[0].enabled)

        # 1
        self.assertEqual(slots[1].index, 1)
        self.assertEqual(slots[1].time, '8:05')
        self.assertEqual(slots[1].weekdays, [0,1,2,3,4]) # Mon-Fri
        self.assertEqual(slots[1].volume, 5)
        self.assertEqual(slots[1].duration, 0)
        self.assertEqual(slots[1].station_data, {
            'title': 'BBC Radio 4',
            'description': 'Intelligent speech',
            'thumbnail_url': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
            'station_url': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
        })
        self.assertTrue(slots[1].enabled)

    def test_set(self):
        alarm, api = get_alarm()

        alarm.slot(2).set(
            time='7:12',
            weekdays=[5, 6], # Sat,Sun
            station_data={
                'title': 'BBC Radio 4',
                'description': 'Intelligent speech',
                'thumbnail_url': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
                'station_url': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
            },
            enabled=True,
        )

        api.set_alarm_info.assert_called_once_with(
            index=2,
            hour=7,
            minute=12,
            week='0x41',
            duration=0,
            volume=5,
            station_data={
                'title': 'BBC Radio 4',
                'description': 'Intelligent speech',
                'thumbnail': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
                'stationurl': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
            }
        )
        api.set_alarm_on_off.assert_called_once_with(2, 'on')

    def test_set_from_playlist(self):
        alarm, api = get_alarm()

        api.get_station_data.return_value = {
            'cpname': 'TuneIn',
            'browsemode': '0',
            'title': 'BBC Radio 4',
            'description': 'Intelligent speech',
            'thumbnail': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
            'stationurl': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
            'timestamp': '2019-01-08T15:21:47Z',
        }

        playlist = [
            type('Item', (object, ), {
                'object_type': 'unsupported_type',
                'object_id': 1,
                'name': 'Some audio'
            }),
            type('Item', (object, ), {
                'object_type': 'tunein_radio',
                'object_id': 20,
                'name': 'Some radio'
            }),
            type('Item', (object, ), {
                'object_type': 'tunein_radio',
                'object_id': 21,
                'name': 'Some other radio'
            })
        ]

        alarm.slot(2).set(
            time='7:12',
            weekdays=[5, 6], # Sat,Sun
            playlist=playlist,
            enabled=True,
        )

        api.set_alarm_info.assert_called_once_with(
            index=2,
            hour=7,
            minute=12,
            week='0x41',
            duration=0,
            volume=5,
            station_data={
                'title': 'BBC Radio 4',
                'description': 'Intelligent speech',
                'thumbnail': 'http://cdn-radiotime-logos.tunein.com/s25419d.png',
                'stationurl': 'http://opml.radiotime.com/Tune.ashx?id=s25419&partnerId=qDDAbg6M&serial=90F1AAD31D82&formats=mp3,wma,aac,qt,hls',
            }
        )
        api.set_alarm_on_off.assert_called_once_with(2, 'on')

    @unittest.mock.patch('datetime.datetime')
    def test_delete(self, dt):
        alarm, api = get_alarm()

        dt.now.return_value = NOW

        slots = alarm.get_slots()
        slots[0].delete()

        self.assertEqual(slots[0].index, 0)
        self.assertEqual(slots[0].time, '15:45')
        self.assertEqual(slots[0].weekdays, [NOW.weekday()])
        self.assertEqual(slots[0].volume, 5)
        self.assertEqual(slots[0].duration, 0)
        self.assertEqual(slots[0].station_data, None)
        self.assertFalse(slots[0].enabled)

        api.del_alarm.assert_called_once_with(0)

    def test_disable(self):
        alarm, api = get_alarm()

        slots = alarm.get_slots()
        slots[0].disable()

        self.assertFalse(slots[0].enabled)
        api.set_alarm_on_off.assert_called_once_with(0, 'off')

    def test_enable(self):
        alarm, api = get_alarm()

        slots = alarm.get_slots()
        slots[0].enable()

        self.assertTrue(slots[0].enabled)
        api.set_alarm_on_off.assert_called_once_with(0, 'on')



import unittest
from unittest.mock import MagicMock
from unittest.mock import call

from samsung_multiroom.service import Item
from samsung_multiroom.service.tunein import TuneInBrowser


def browser_main_return_value():
    return [
        {
            '@type': '0',
            'title': 'Favorites',
            'contentid': '0',
        },
        {
            '@type': '0',
            'title': 'Local Radio',
            'contentid': '1',
        },
        {
            '@type': '0',
            'title': 'Recents',
            'contentid': '2',
        },
        {
            '@type': '0',
            'title': 'Trending',
            'contentid': '3',
        },
        {
            '@type': '0',
            'title': 'By Language',
            'contentid': '10',
        }
    ]


def get_select_radio_list_side_effect(content_id, start_index, list_count):
    if content_id == '10':
        return [
            {
                '@type': '0',
                'title': 'Arabic',
                'contentid': '4',
            },
            {
                '@type': '0',
                'title': 'English',
                'contentid': '24',
            },
            {
                '@type': '0',
                'title': 'Finnish',
                'contentid': '28',
            },
            {
                '@type': '0',
                'title': 'French',
                'contentid': '29',
            },
        ]
    if content_id == '24':
        return [
            {
                '@type': '0',
                'title': 'Music',
                'contentid': '0',
            },
            {
                '@type': '0',
                'title': 'Talk',
                'contentid': '1',
            },
        ]
    if content_id == '0':
        return [
            {
                '@type': '2',
                'thumbnail': 'http://cdn-profiles.tunein.com/s297990/images/logot.png',
                'description': 'MSNBC Live with Velshi & Ruhle',
                'mediaid': 's297990',
                'title': 'MSNBC',
                'contentid': '0',
            },
            {
                '@type': '2',
                'thumbnail': 'http://cdn-radiotime-logos.tunein.com/s24940t.png',
                'description': 'Amazing music. Played by an amazing line up.',
                'mediaid': 's24940',
                'title': 'BBC1',
                'contentid': '1',
            }
        ]


class TestTuneInBrowser(unittest.TestCase):

    @unittest.mock.patch('inspect.signature')
    def test_browse_from_root(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.browse_main.return_value = browser_main_return_value()

        browser = TuneInBrowser(api)
        browser = browser.browse()

        api.browse_main.assert_called_once_with(start_index=0, list_count=30)

        self.assertEqual(browser.get_path(), '/')
        self.assertEqual(len(browser), 5)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].name, 'Favorites')
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'container')

    @unittest.mock.patch('inspect.signature')
    def test_browse_full_level(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.browse_main.return_value = browser_main_return_value()
        api.get_select_radio_list.side_effect = get_select_radio_list_side_effect

        browser = TuneInBrowser(api)
        browser = browser.browse('/By Language/English/Music')

        api.browse_main.assert_called_once_with(start_index=0, list_count=30)
        api.get_select_radio_list.assert_has_calls([
            call(content_id='10', start_index=0, list_count=30),
            call(content_id='24', start_index=0, list_count=30),
            call(content_id='0', start_index=0, list_count=30),
        ])

        self.assertEqual(browser.get_path(), '/By Language/English/Music')
        self.assertEqual(len(browser), 2)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].name, 'MSNBC')
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'tunein_radio')

    @unittest.mock.patch('inspect.signature')
    def test_browse_relative(self, signature):
        signature.side_effect = [
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
            type('signature', (object, ), {'parameters': {'content_id': None, 'start_index': None, 'list_count': None}}),
        ]

        api = MagicMock()
        api.browse_main.return_value = browser_main_return_value()
        api.get_select_radio_list.side_effect = get_select_radio_list_side_effect

        browser = TuneInBrowser(api)
        browser = browser.browse('/By Language/English').browse('Music')

        api.browse_main.assert_called_once_with(start_index=0, list_count=30)
        api.get_select_radio_list.assert_has_calls([
            call(content_id='10', start_index=0, list_count=30),
            call(content_id='24', start_index=0, list_count=30),
            call(content_id='0', start_index=0, list_count=30),
        ])

        self.assertEqual(browser.get_path(), '/By Language/English/Music')
        self.assertEqual(len(browser), 2)
        self.assertIsInstance(browser[0], Item)
        self.assertEqual(browser[0].name, 'MSNBC')
        self.assertEqual(browser[0].object_id, '0')
        self.assertEqual(browser[0].object_type, 'tunein_radio')



"""Player allows playback control depending on selected source."""
import abc

# repeat mode constants
REPEAT_ONE = 'one'
REPEAT_ALL = 'all'
REPEAT_OFF = 'off'


class Player(metaclass=abc.ABCMeta):
    """Player interface to control playback functions."""

    @abc.abstractmethod
    def play(self, playlist):
        """
        Enqueue and play a playlist.

        Player may choose to not play the playlist if it's not compatible with this player. For instance you can't
        play DLNA source tracks using TuneIn player. If player is unable to play the playlist it must return False.

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def jump(self, time):
        """
        Advance current playback to specific time.

        :param time: Time from the beginning of the track in seconds
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def resume(self):
        """Play/resume current track."""
        raise NotImplementedError()

    @abc.abstractmethod
    def stop(self):
        """Stop current track and reset position to the beginning."""
        raise NotImplementedError()

    @abc.abstractmethod
    def pause(self):
        """Pause current track and retain position."""
        raise NotImplementedError()

    @abc.abstractmethod
    def next(self):
        """Play next track in the queue."""
        raise NotImplementedError()

    @abc.abstractmethod
    def previous(self):
        """Play previous track in the queue."""
        raise NotImplementedError()

    @abc.abstractmethod
    def repeat(self, mode):
        """
        Set playback repeat mode.

        :param mode: one of REPEAT_* constants
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_repeat(self):
        """
        Get playback repeat mode.

        :returns: one of REPEAT_* constants
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def is_supported(self, function, submode=None):
        """
        Check if this player supports function/submode.

        :returns: Boolean True if function/submode is supported
        """
        raise NotImplementedError()


class PlayerOperator(Player):
    """Select the right player for the current source."""

    def __init__(self, api, players=None):
        """
        Initialise player operator.

        :param api: SamsungMultiroomApi instance
        :param players: List of Player instances
        """
        self._api = api
        self._players = []

        for player in (players or []):
            self.add_player(player)

    def add_player(self, player):
        """
        Add player.

        :param player: Player instance
        """
        if not isinstance(player, Player):
            raise ValueError('Player must be a subclass of Player class')

        self._players.append(player)

    def get_player(self):
        """
        Get currently active player based on selected source.

        :returns: Player instance
        """
        func = self._api.get_func()

        function = func['function']
        submode = func['submode']

        for player in self._players:
            if player.is_supported(function, submode):
                return player

        return NullPlayer()

    def play(self, playlist):
        """
        Find a suitable player and play a playlist.

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        for player in self._players:
            if player.play(playlist):
                return True

        return False

    def jump(self, time):
        """
        Advance current playback to specific time.

        :param time: Time from the beginning of the track in seconds
        """
        self.get_player().jump(time)

    def resume(self):
        """Play/resume current track."""
        self.get_player().resume()

    def stop(self):
        """Stop current track and reset position to the beginning."""
        self.get_player().stop()

    def pause(self):
        """Pause current track and retain position."""
        self.get_player().pause()

    def next(self):
        """Play next track in the queue."""
        self.get_player().next()

    def previous(self):
        """Play previous track in the queue."""
        self.get_player().previous()

    def repeat(self, mode):
        """
        Set playback repeat mode.

        :param mode: one of REPEAT_* constants
        """
        self.get_player().repeat(mode)

    def get_repeat(self):
        """
        Get playback repeat mode.

        :returns: one of REPEAT_* constants
        """
        return self.get_player().get_repeat()

    def get_current_track(self):
        """
        Get current track info.

        :returns: Track instance, or None if unavailable
        """
        return self.get_player().get_current_track()

    def is_supported(self, function, submode=None):
        """
        Check if this player supports function/submode.

        :returns: Boolean True if function/submode is supported
        """
        for player in self._players:
            if player.is_supported(function, submode):
                return True

        return False


class NullPlayer(Player):
    """Catch all player if no others supported current function."""

    def play(self, playlist):
        """
        Null player is unable to play anything.

        :param playlist: Playlist instance
        :returns: Boolean False
        """
        return False

    def jump(self, time):
        """Do nothing."""

    def resume(self):
        """Do nothing."""

    def stop(self):
        """Do nothing."""

    def pause(self):
        """Do nothing."""

    def next(self):
        """Do nothing."""

    def previous(self):
        """Do nothing."""

    def repeat(self, mode):
        """Do nothing."""

    def get_repeat(self):
        """Always off."""
        return REPEAT_OFF

    def get_current_track(self):
        """
        Do nothing.

        :returns: None
        """
        return None

    def is_supported(self, function, submode=None):
        """
        Check if this player supports function/submode.

        :returns: Boolean True
        """
        return True


class Track:
    """Defines a media track on the playlist."""

    def __init__(self, title, artist, album, duration, position, thumbnail_url, metadata=None):
        self._title = title
        self._artist = artist
        self._album = album
        self._duration = duration
        self._position = position
        self._thumbnail_url = thumbnail_url
        self._metadata = metadata or {}

    @property
    def title(self):
        """
        :returns: Title of the current track
        """
        return self._title

    @property
    def artist(self):
        """
        :returns: Artist of the current track
        """
        return self._artist

    @property
    def album(self):
        """
        :returns: Album title of the current track
        """
        return self._album

    @property
    def duration(self):
        """
        :returns: Duration in seconds
        """
        return self._duration

    @property
    def position(self):
        """
        :returns: Current playback position in seconds
        """
        return self._position

    @property
    def thumbnail_url(self):
        """
        :returns: URL of the track thumbnail
        """
        return self._thumbnail_url

    def __getattr__(self, name):
        """
        :returns: Metadata item value
        """
        if name in self._metadata:
            return self._metadata[name]

        return None


def init_track_kwargs(object_type):
    """
    :returns: kwargs dict fro Track initialisation
    """
    return {
        'title': None,
        'artist': None,
        'album': None,
        'duration': None,
        'position': None,
        'thumbnail_url': None,
        'metadata': {
            'object_id': None,
            'object_type': object_type,
        }
    }



"""Abstract media browser and item."""
import abc


class Browser(metaclass=abc.ABCMeta):
    """
    Abstract media browser.

    Implementations should be immutable.
    """

    def __init__(self, path=None, items=None):
        """
        :param path: Path used to list items
        :param items: Items listed
        """
        self._path = path
        self._items = items.copy() if items else []

    @abc.abstractmethod
    def get_name(self):
        """
        :returns: Name of the browser
        """
        raise NotImplementedError()

    def get_path(self):
        """
        :returns: Path used to list items
        """
        return self._path

    @abc.abstractmethod
    def browse(self, path=None):
        """
        List items based on path.

        :param path: Forward slash separated string of folders e.g. /NAS/Music/By Folder/Air/Moon Safari/CD1
        :returns: Browser instance with listed items
        """
        raise NotImplementedError()

    def _merge_path(self, new_path):
        folders = path_to_folders(new_path)

        if folders[0] is not None:
            folders = path_to_folders(self.get_path()) + folders

        return folders_to_path(folders)

    def _get_items(self):
        return self._items

    def _get_initial_items(self, folders):
        """
        :returns: initial items to use, for root search return empty list
        """
        if folders[0] is None:
            return []

        return self._get_items()

    def _get_initial_depth(self, folders):
        """
        :returns: depth of the current browse, taking into account relative/absolute browsing
        """
        if folders[0] is None:
            return 0

        return len(path_to_folders(self._path))

    def __getitem__(self, i):
        return self._items[i]

    def __len__(self):
        return len(self._items)

    def __iter__(self):
        return iter(self._items)


class Item:
    """
    Immutable browser item.
    """

    def __init__(self, object_id, object_type, name, metadata=None):
        self._object_id = object_id
        self._object_type = object_type
        self._name = name
        self._metadata = metadata or {}

    @property
    def object_id(self):
        """
        :returns: Object id
        """
        return self._object_id

    @property
    def object_type(self):
        """
        :returns: Object type
        """
        return self._object_type

    @property
    def name(self):
        """
        :returns: Item name
        """
        return self._name

    @property
    def title(self):
        """
        :returns: Item title, equivallent of name
        """
        if 'title' in self._metadata:
            return self._metadata['title']

        return self.name

    @property
    def metadata(self):
        """
        :returns: Item metadata specific to item type
        """
        return self._metadata

    def __getattr__(self, name):
        """
        Look up name in metadata.

        :returns: Metadata value
        """
        if name in self._metadata:
            return self._metadata[name]

        return None


def path_to_folders(path):
    """
    Convert path to a standarised list of folder.

    If path is prepended with /, first element on the list will be special folder None denoting root.

    :param path: None or path string with folders '/' separated, e.g. /NAS/Music/By Folder/folder1/folder2
    :returns: List of folders e.g. [None, 'NAS', 'Music', 'By Folder', 'folder1', 'folder2']
    """
    if path in [None, '', '/']:
        path = '/'

    is_from_root = (path[0] == '/')

    path = path.strip(' /').split('/')
    path = [p for p in path if p != '']

    if is_from_root:
        path.insert(0, None)

    return path


def folders_to_path(folders):
    """
    Convert list of folders to a path.

    :param folders: List of folders, None denotes a root folder e.g. [None, 'NAS', 'Music', 'By Folder', 'folder1']
    :returns: path string e.g. /NAS/Music/By Folder/folder1
    """
    if not folders:
        return '/'

    if folders[0] is None:
        folders[0] = ''

    return '/'.join(folders) or '/'



"""TuneIn service player."""
from ...api import paginator
from ..player import REPEAT_OFF
from ..player import Player
from ..player import Track
from ..player import init_track_kwargs


class TuneInPlayer(Player):
    """Controls player in WIFI+ TuneIn radio mode."""

    def __init__(self, api):
        self._api = api

    def play(self, playlist):
        """
        Play first radio from a playlist.

        Playlist items must be an object with following attributes:
        - object_id - object id
        - object_type - must be 'tunein_radio'
        - title - radio name

        :param playlist: Iterable returning player combatible objects
        :returns: True if playlist was accepted, False otherwise
        """
        for radio in playlist:
            if radio.object_type not in ['tunein_radio']:
                continue

            self._api.set_play_select(radio.object_id)
            return True

        return False

    def jump(self, time):
        """
        Not supported for radios.

        :param time: Time from the beginning of the track in seconds
        """

    def resume(self):
        """Play/resume current track."""
        self._api.set_select_radio()

    def stop(self):
        """Stop current radio."""
        raise NotImplementedError()

    def pause(self):
        """Pause current radio."""
        self._api.set_playback_control('pause')

    def next(self):
        """Play next radio from the preset list."""
        self._previous_next_preset(1)

    def previous(self):
        """Play previous radio from the preset list."""
        self._previous_next_preset(-1)

    def repeat(self, mode):
        """Not supported for radio."""

    def get_repeat(self):
        """Not supported for radio."""
        return REPEAT_OFF

    def get_current_track(self):
        """
        Get current radio info.

        :returns: Track instance, or None if unavailable
        """
        radio_info = self._api.get_radio_info()

        track_kwargs = init_track_kwargs('tunein_radio')

        if 'title' in radio_info:
            track_kwargs['artist'] = radio_info['title']
        if 'description' in radio_info:
            track_kwargs['title'] = radio_info['description']
        if 'thumbnail' in radio_info and 'http' in radio_info['thumbnail']:
            track_kwargs['thumbnail_url'] = radio_info['thumbnail']

        return Track(**track_kwargs)

    def is_supported(self, function, submode=None):
        """
        Check if this player supports function/submode.

        :returns: Boolean True if function/submode is supported
        """
        return function == 'wifi' and submode == 'cp'

    def _previous_next_preset(self, direction):
        if direction not in [-1, 1]:
            raise ValueError('Direction must be either 1 or -1')

        presets = list(paginator(self._api.get_preset_list, 0, 30))
        presets_count = len(presets)

        if presets_count <= 1:
            return

        # locate current mediaid on the preset list
        radio_info = self._api.get_radio_info()

        current_preset_index = -direction
        if 'presetindex' in radio_info and radio_info['presetindex'] is not None:
            current_preset_index = int(radio_info['presetindex'])

        kind_preset_type = {
            'speaker': 1,
            'my': 0,
        }

        preset_index = (presets_count + max(0, current_preset_index) + direction) % presets_count
        preset = presets[preset_index]
        preset_type = kind_preset_type[preset['kind']]

        self._api.set_play_preset(preset_type, preset_index)
        self._api.set_select_radio()



"""Low level api to communicate with samsung multiroom speaker."""
import logging
import urllib.parse
import uuid

import requests

from .api_response import ApiResponse
from .api_stream import ApiStream

METHOD_GET = 'get'

COMMAND_UIC = 'UIC'
COMMAND_CPM = 'CPM'

_LOGGER = logging.getLogger(__name__)


class SamsungMultiroomApiException(Exception):
    """Generic API exception."""


class SamsungMultiroomApi:
    """
    Samsung Multiroom Api implementation.

    Contains non-inclusive list of API calls you can make to control the speaker.
    """

    def __init__(self, ip_address, port=55001, timeout=5):
        """
        Initialise endpoint.

        :param ip_address: IP address of the speaker to connect to
        :param port: Port to use, defaults to 55001
        :param timeout: Timeout in seconds
        """
        self._ip_address = ip_address
        self._port = port
        self._endpoint = 'http://{0}:{1}'.format(ip_address, port)
        self._timeout = timeout
        self._uuid = str(uuid.uuid1())

    @property
    def ip_address(self):
        """
        :returns: IP address of the host
        """
        return self._ip_address

    @property
    def port(self):
        """
        :returns: Port to use
        """
        return self._port

    def request(self, method, command, payload):
        """
        Makes a request to a configured endpoint.

        :param method: HTTP method to use
        :param command: UIC|CPM
        :param payload: XML string with payload
        :returns: Response dict
        :raises: ValueError
        :raises: SamsungMultiroomApiException
        """
        if method not in [METHOD_GET]:
            raise ValueError('Invalid method {0}, must be one of METHOD_* constants'.format(method))

        if command not in [COMMAND_UIC, COMMAND_CPM]:
            raise ValueError('Invalid command {0}, must be one of COMMAND_* constants'.format(method))

        url = '{0}/{1}?cmd={2}'.format(self._endpoint, command, urllib.parse.quote(payload))
        headers = {
            'mobileUUID': self._uuid,
            'mobileName': 'Wireless Audio',
            'mobileVersion': '1.0',
        }

        try:
            _LOGGER.debug('Request %s. Raw payload %s', url, payload)
            response = requests.get(url, headers=headers, timeout=self._timeout)

            return self._parse_response_text(response.text)
        except requests.exceptions.RequestException:
            _LOGGER.error('Request %s failed', url, exc_info=1)
            raise SamsungMultiroomApiException('Request {0} failed'.format(url))

    def _parse_response_text(self, response_text):
        _LOGGER.debug('Response %s', response_text)

        response = ApiResponse(response_text)

        if not response.success:
            raise SamsungMultiroomApiException('Received invalid response {0}'.format(response.raw))

        return response.data

    def get(self, command, action, params=None):
        """
        Generic GET request.

        :param command: COMMAND_* constant
        :param action: Get action name to execute e.g. GetVolume
        :param params: List of tuples (name, value, (optional) type hint str/dec/cdata)
        :returns dict
        """
        return self.request(METHOD_GET, command, format_payload(action, params))

    def get_speaker_name(self):
        """
        :returns: Speaker name
        """
        return self.get(COMMAND_UIC, 'GetSpkName')['spkname']

    def set_speaker_name(self, name):
        """
        Set speaker name.

        :param name: new speaker name
        """
        self.get(COMMAND_UIC, 'SetSpkName', [('spkname', name, 'cdata')])

    def get_main_info(self):
        """
        Get main information about speaker.

        :returns: Dict:
            - party - off
            - partymain - None
            - grouptype - one of M, S, N (Master, Slave, None?)
            - groupmainip - if speaker is in group, this is IP address of the master speaker
            - groupmainmacaddr - if speaker is in group, this is MAC address of the master speaker
            - spkmacaddr - this speaker's MAC address
            - spkmodelname - this speaker's model
            - groupmode - aasync, none
            - channeltype - front, invalid
            - channelvolume - 0
            - multichinfo - on
            - groupspknum - total number of speakers in the group
            - dfsstatus - dfsoff
            - protocolver - 2.3
            - btmacaddr - bluetooth MAC address
        """
        path = '/{0}?cmd={1}'.format(COMMAND_UIC, urllib.parse.quote(format_payload('GetMainInfo')))

        stream = ApiStream(self._ip_address, self._port, self._timeout)

        # Speaker sends two http responses for this request, latter one contains correct payload. We attempt
        # to fetch both responses and read/parse both.
        for response in stream.open(path):
            if not response.success:
                stream.close()
                _LOGGER.error('Request http://%s:%s%s failed', self._ip_address, self._port, path, exc_info=1)
                raise SamsungMultiroomApiException('Request http://{0}:{1}{2} failed'.format(
                    self._ip_address, self._port, path))

            if response.name == 'MainInfo':
                stream.close()
                return response.data

        _LOGGER.error('Request http://%s:%s%s failed', self._ip_address, self._port, path, exc_info=1)
        raise SamsungMultiroomApiException('Request http://{0}:{1}{2} failed'.format(
            self._ip_address, self._port, path))

    def get_volume(self):
        """
        Get current volume level between 0 and 100.

        :returns: int - volume level
        """
        return int(self.get(COMMAND_UIC, 'GetVolume')['volume'])

    def set_volume(self, volume):
        """
        Set speaker volume level.

        :param volume: Volume level between 0 and 100
        """
        self.get(COMMAND_UIC, 'SetVolume', [('volume', int(volume))])

    def get_mute(self):
        """
        Get mute state of the speaker.

        :returns: boolean True if muted
        """
        return on_off_bool(self.get(COMMAND_UIC, 'GetMute')['mute'])

    def set_mute(self, mute):
        """
        Mute/unmute the speaker.

        :param mute: boolean True to mute
        """
        self.get(COMMAND_UIC, 'SetMute', [('mute', bool_on_off(mute))])

    def get_func(self):
        """
        Retrieve current source for the speaker.

        :returns: Dict with
            - function - aux|bt|hdmi|optical|soundshare|wifi
            - submode - dlna|cp|?
        """
        return self.get(COMMAND_UIC, 'GetFunc')

    def set_func(self, function):
        """
        Set the source for the speaker

        :param function: aux|bt|hdmi|optical|soundshare|wifi
        """
        self.get(COMMAND_UIC, 'SetFunc', [('function', function)])

    def get_shuffle_mode(self):
        """
        Retrieve currently set shuffle mode.

        :returns: Boolean True if shuffle mode is enabled
        """
        return on_off_bool(self.get(COMMAND_UIC, 'GetShuffleMode')['shuffle'])

    def set_shuffle_mode(self, shuffle_mode):
        """
        Enable/disable shuffle mode of the playlist.

        :param shuffle_mode: boolean
        """
        self.get(COMMAND_UIC, 'SetShuffleMode', [('shufflemode', bool_on_off(shuffle_mode))])

    def set_trick_mode(self, trick_mode):
        """
        Move to next/previous track on the playlist.

        :param trick_mode: previous|next
        """
        if trick_mode not in ['previous', 'next']:
            raise ValueError('Trick mode must one of: previous, next')

        self.get(COMMAND_UIC, 'SetTrickMode', [('trickmode', trick_mode)])

    def set_playback_control(self, playback_control):
        """
        Pause/resume current playlist.

        :param playback_control: resume|pause|play
        """
        if playback_control not in ['resume', 'pause']:
            raise ValueError('Playback control must be one of: resume, pause')

        self.get(COMMAND_UIC, 'SetPlaybackControl', [('playbackcontrol', playback_control)])

    def get_music_info(self):
        """
        Get detailed information about current track on the playlist.

        :returns: Dict with
            - device_udn - source device unique identifier
            - playertype - allshare
            - playbacktype - folder|playlist
            - sourcename
            - parentid
            - parentid2
            - objectid
            - title
            - artist
            - album
            - thumbnail
            - timelength - HH:MM:SS.uuu format
            - playtime - in microseconds
            - seek - enable|?
            - pause - enable|?
        """
        return self.get(COMMAND_UIC, 'GetMusicInfo')

    def get_play_status(self):
        """
        Get information about play/pause status for the playlist.

        :returns: Dict with
            - function - wifi|?
            - submode - dlna|cp
            - playstatus - (optional) play|pause
        """
        return self.get(COMMAND_UIC, 'GetPlayStatus')

    def set_search_time(self, play_time):
        """
        Set current track to play from a specific time.

        API call will fail unless track is currently played.

        Setting play time further than length of the track ends the current track and plays next from the beginning.

        :play_time: play time in seconds
        """
        self.get(COMMAND_UIC, 'SetSearchTime', [('playtime', int(play_time))])

    def get_preset_list(self, start_index, list_count):
        """
        Get list of predefined radios.

        :param start_index: Starting position to retrieve
        :param list_count: Total number of items to retrieve
        :returns: List of dicts
            - title
            - description
            - thumbnail
            - contentid
            - mediaid
        """
        response = self.get(COMMAND_CPM, 'GetPresetList', [
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ])

        if not int(response['listcount']):
            return []

        return response_list(response['presetlist']['preset'])

    def get_radio_info(self):
        """
        Retrieves currently selected radio info and play status.

        :returns: Dict
            - cpname - TuneIn|Unknown|?
            - playstatus - stop|play
            Optionally:
            - root - preset category
            - presetindex -  position on the preset list
            - title - name of the radio
            - description - description of the radio, can include the currently played song
            - thumbnail - URL of the radio thumbnail image
            - mediaid
            - allowfeedback
            - timestamp - in ISO 8601 format
            - noqueue
        """
        return self.get(COMMAND_CPM, 'GetRadioInfo')

    def set_play_preset(self, preset_type, preset_index):
        """
        Select radio of a particular index.

        Combine this with set_select_radio to play selected radio.

        :param preset_type: 1 - speaker, 0 - my
        :param preset_index: Index of get preset list
        """
        self.get(COMMAND_CPM, 'SetPlayPreset', [
            ('presettype', int(preset_type)),
            ('presetindex', int(preset_index)),
        ])

    def set_select_radio(self):
        """
        Play selected preset preset.

        Precede this with set_play_preset.
        """
        self.get(COMMAND_CPM, 'SetSelectRadio')

    def get_dms_list(self, start_index, list_count):
        """
        Retrieve list of DLNA compatible devices to use as a media source.

        :param start_index:
        :param list_count:
        :returns: List of DLNA devices dicts
            - @device_id - likely a sequential id
            - dmsid - device udn e.g. uuid:00113249-398f-0011-8f39-8f3949321100
            - dmsname - device name e.g. nas
            - devicetype - e.g. network
            - thumbnail_PNG_LRG - thumbnail url
            - thumbnail_JPG_LRG - thumbnail url
            - thumbnail_PNG_SM - thumbnail url
            - thumbnail_JPG_SM - thumbnail url
        """
        response = self.get(COMMAND_UIC, 'GetDmsList', [
            ('liststartindex', int(start_index)),
            ('listcount', int(list_count)),
        ])

        if not int(response['listcount']):
            return []

        return response_list(response['dmslist']['dms'])

    def pc_get_music_list_by_category(self, device_udn, start_index, list_count):
        """
        Browse containers at the root of the DLNA device.

        :param device_udn: dmsid returned by get_dms_list()
        :param start_index:
        :param list_count:
        :returns: List of category dicts
            - @object_id - use it to browse into this container
            - type - usually CONTAINER
            - playindex - -1 for CONTAINERs
            - name - folder name
            - title - None
            - artist - None
            - album - None
            - thumbnail - None
            - timelength - None
            - device_udn -
        """
        response = self.get(COMMAND_UIC, 'PCGetMusicListByCategory', [
            ('device_udn', device_udn),
            ('filter', 'folder'),
            ('categoryid', 'folder'),
            ('liststartindex', int(start_index)),
            ('listcount', int(list_count)),
        ])

        if not int(response['listcount']):
            return []

        return response_list(response['musiclist']['music'])

    def pc_get_music_list_by_id(self, device_udn, parent_id, start_index, list_count):
        """
        Browse containers/audio items in the container of the DLNA device.

        :param device_udn: dmsid returned by get_dms_list()
        :param parent_id: object_id as returned by pc_get_music_list_by_category or this method
        :param start_index:
        :param list_count:
        :returns: List of containers/audio items dicts
            - @object_id - e.g. 22$@52941
            - type - CONTAINER|AUDIO
            - playindex - -1 for CONTAINER, sequential for AUDIO
            - name - folder/file name
            - title
            - artist
            - album
            - thumbnail - URL
            - timelength - HH:MM:SS.xxx format
            - device_udn -
        """
        response = self.get(COMMAND_UIC, 'PCGetMusicListByID', [
            ('device_udn', device_udn),
            ('filter', 'folder'),
            ('parentid', str(parent_id)),
            ('liststartindex', int(start_index)),
            ('listcount', int(list_count)),
        ])

        if not int(response['listcount']):
            return []

        return response_list(response['musiclist']['music'])

    def set_playlist_playback_control(self, items):
        """
        Create a playlist and playback.

        Use pc_get_music_list_by_id() to fetch item information required for playlist item.

        :param items: List of dicts:
            - device_udn
            - object_id
            - title - song title
            - artist - song artist
            - thumbnail - URL
        """
        params = [
            ('playbackcontrol', 'play'),
            ('playertype', 'allshare'),
            ('sourcename', '', 'cdata'),
            ('playindex', 0),
            ('playtime', 0),
            ('totalobjectcount', len(items)),
        ]

        for item in items:
            if 'title' not in item:
                item['title'] = 'Unknown'
            if 'artist' not in item:
                item['artist'] = 'Unknown'
            if 'thumbnail' not in item:
                item['thumbnail'] = ''

            params.append(('device_udn', item['device_udn']))
            params.append(('objectid', item['object_id']))
            params.append(('songtitle', item['title'], 'cdata'))
            params.append(('thumbnail', item['thumbnail'], 'cdata'))
            params.append(('artist', item['artist'], 'cdata'))

        self.get(COMMAND_UIC, 'SetPlaylistPlaybackControl', params)

    def browse_main(self, start_index, list_count):
        """
        Browse radios from the root.

        :param start_index:
        :param list_count:
        :returns: radio list item dict:
            folder:
            - @type - 0 - folder, 2 - radio
            - title - name of the folder
            - contentid - pass to get_select_radio_list() to browse into this folder
            radio:
            - @type - 0 - folder, 2 - radio
            - title - name of the radio
            - description - radio description
            - mediaid
            - thumbnail - URL
            - contentid
        """
        params = [
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'BrowseMain', params)

        if not int(response['listcount']):
            return []

        return response_list(response['menulist']['menuitem'])

    def get_select_radio_list(self, content_id, start_index, list_count):
        """
        Browse specific radio folder.

        Note: you can't browse arbitrary folder. In order to browse a folder you had to navigate to it in a previous
        browse_main() or get_select_radio_list() call. This limitation is imposed by the speaker.

        :param content_id: contentid as returned from browse_main() or get_select_radio_list()
        :param start_index:
        :param list_count:
        :returns: see browse_main()
        """
        params = [
            ('contentid', int(content_id)),
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'GetSelectRadioList', params)

        if not int(response['listcount']):
            return []

        return response_list(response['menulist']['menuitem'])

    def get_current_radio_list(self, start_index, list_count):
        """
        Browse previously browsed radio folder.

        :param start_index:
        :param list_count:
        :returns: see browse_main()
        """
        params = [
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'GetCurrentRadioList', params)

        if not int(response['listcount']):
            return []

        return response_list(response['menulist']['menuitem'])

    def get_upper_radio_list(self, start_index, list_count):
        """
        Browse parent of a browsed radio folder.

        :param start_index:
        :param list_count:
        :returns: see browse_main()
        """
        params = [
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'GetUpperRadioList', params)

        if not int(response['listcount']):
            return []

        return response_list(response['menulist']['menuitem'])

    def set_play_select(self, content_ids):
        """
        Plays selected radio or app item.

        :param content_ids: Content id as returned by get_upper_radio_list(), get_select_radio_list() or
            get_current_radio_list(), or list of content ids
        """
        if not isinstance(content_ids, list):
            content_ids = [content_ids]

        if len(content_ids) > 1:
            params = [('selectitemids', [int(id) for id in content_ids])]
        elif content_ids:
            params = [('selectitemid', int(content_ids[0]))]
        else:
            params = [('selectitemid', '')]

        self.get(COMMAND_CPM, 'SetPlaySelect', params)

    def get_station_data(self, content_id):
        """
        Get radio station data.

        :param content_id: Content id as returned by get_upper_radio_list(), get_select_radio_list() or
            get_current_radio_list()
        :returns: Station data dict:
            - cpname - likely TuneIn
            - title
            - description
            - thumbnail
            - stationurl
            - browsemode
            - timestamp
        """
        params = [('selectitemid', int(content_id))]

        return self.get(COMMAND_CPM, 'GetStationData', params)

    def get_7band_eq_list(self):
        """
        Retrieve equalizer presets.

        :returns: List of preset dicts with following attributes:
            - @index
            - presetindex
            - presetname
        """
        response = self.get(COMMAND_UIC, 'Get7BandEQList')

        return response_list(response['presetlist']['preset'])

    def get_current_eq_mode(self):
        """
        Retrieve current equalizer settings.

        :returns: Preset dict with following attributes:
            - presetindex
            - presetname
            - eqvalue1
            - eqvalue2
            - eqvalue3
            - eqvalue4
            - eqvalue5
            - eqvalue6
            - eqvalue7
        """
        response = self.get(COMMAND_UIC, 'GetCurrentEQMode')

        return response

    def set_7band_eq_value(self, preset_index, values):
        """
        Set preset's equalizer settings.

        Note, this doesn't overwrite preset settings, this method only sets those values temporarily. To overwrite use
        reset_7band_eq_value() method.

        :param preset_index:
        :param values: List of 7 integers ranging between -6 and 6
        """
        params = [('presetindex', int(preset_index))]

        for i, value in enumerate(values):
            params.append(('eqvalue' + str(i + 1), int(value)))

        self.get(COMMAND_UIC, 'Set7bandEQValue', params)

    def set_7band_eq_mode(self, preset_index):
        """
        Switch equalizer to a predefined preset.

        :param preset_index:
        """
        params = [('presetindex', int(preset_index))]

        self.get(COMMAND_UIC, 'Set7bandEQMode', params)

    def reset_7band_eq_value(self, preset_index, values):
        """
        Overwrite preset's equalizer settings.

        :param preset_index:
        :param values: List of 7 integers ranging between -6 and 6
        """
        params = [('presetindex', int(preset_index))]

        for i, value in enumerate(values):
            params.append(('eqvalue' + str(i + 1), int(value)))

        self.get(COMMAND_UIC, 'Reset7bandEQValue', params)

    def del_custom_eq_mode(self, preset_index):
        """
        Delete custom preset.

        Note, you cannot delete predefined presets with indices between 0 and 3 inculsive.

        :param preset_index:
        """
        params = [('presetindex', int(preset_index))]

        self.get(COMMAND_UIC, 'DelCustomEQMode', params)

    def add_custom_eq_mode(self, preset_index, preset_name):
        """
        Creates a new custom preset, using currently set equilizer values.

        Use set_7band_eq_value() or set_7band_eq_mode() to set equilizer values.

        It also allows to overwrite existing custom preset with current equilizer values.

        :param presetindex:
        :param presetname:
        """
        params = [('presetindex', int(preset_index)), ('presetname', preset_name)]

        self.get(COMMAND_UIC, 'AddCustomEQMode', params)

    def set_speaker_time(self, datetime):
        """
        Set speaker's internal time

        :param datetime: Datetime object e.g. datetime.datetime.now()
        """
        params = [
            ('year', datetime.year),
            ('month', datetime.month),
            ('day', datetime.day),
            ('hour', datetime.hour),
            ('min', datetime.minute),
            ('sec', datetime.second),
        ]

        self.get(COMMAND_UIC, 'SetSpeakerTime', params)

    def get_sleep_timer(self):
        """
        Get sleep timer settings
        :returns: Timer settings dict with following attributes:
            - sleepoption - off|start
            - sleeptime - remaining time in seconds
        """
        return self.get(COMMAND_UIC, 'GetSleepTimer')

    def set_sleep_timer(self, option, time):
        """
        Put speaker into sleep mode after specific time
        :param option: off|start
        :param time: delay in seconds
        """
        params = [
            ('option', option),
            ('sleeptime', int(time)),
        ]

        self.get(COMMAND_UIC, 'SetSleepTimer', params)

    def get_alarm_info(self):
        """
        Get list of set alarms.

        :returns: List of dicts:
            - @index - alarm id
            - hour - hour part of alarm
            - min - minutes part of alarm
            - week - hex of days flags Sun Mon Tue Wed Thu Fri Sat, e.g. for weekdays 00111110 - 0x3E
            - volume - volume of alarm
            - title - radio station title
            - description - radio station description
            - thumbnail - radio station thumbnail
            - stationurl - radio station URL
            - set - on|off whether alarm is active or not
            - soundenable - on|off whether predefined sound is used or not
            - sound - on|off whether predefined sound is used or not
            - alarmsoundname - name of predefined alarm sound names as returned by get_alarm_sound_list()
            - duration - duration of alarm in seconds
        """
        response = self.get(COMMAND_UIC, 'GetAlarmInfo')

        return response_list(response['alarmList']['alarm'])

    def set_alarm_on_off(self, index, alarm):
        """
        Enable/disable alarm.

        :param index: Alarm index
        :param alarm: on|off
        """
        params = [
            ('index', int(index)),
            ('alarm', alarm),
        ]

        self.get(COMMAND_UIC, 'SetAlarmOnOff', params)

    def get_alarm_sound_list(self):
        """
        Get list of predefined alarm sounds to use for alarm.

        :returns: List of dicts:
            - @index - alarm sound index
            - alarsoundindex - (note misspelling) alarm sound index
            - alarmsoundname - alarm sound name
        """
        response = self.get(COMMAND_UIC, 'GetAlarmSoundList')

        return response_list(response['alarmlist']['alarmsound'])

    def set_alarm_info(self, index, hour, minute, week, duration, volume, station_data):
        """
        Create alarm.

        Note, you can only create 3 alarms with indices 0, 1, and 2.

        :param index: Alarm index
        :param hour: Alarm hour
        :param minute: Alarm minute
        :param week: hex of days flags Sun Mon Tue Wed Thu Fri Sat, e.g. for weekdays 00111110 - 0x3E
        :param duration: Alarm duration in seconds
        :param volume: Alarm volume 0-100
        :param station_data: Dict as returned by get_station_data
            - title
            - description
            - thumbnail
            - stationurl
        """
        params = [
            ('index', int(index)),
            ('hour', int(hour)),
            ('min', int(minute)),
            ('week', hex(int(week, 16))),
            ('volume', int(volume)),
            ('title', station_data['title'], 'cdata'),
            ('description', station_data['description'], 'cdata'),
            ('thumbnail', station_data['thumbnail'], 'cdata'),
            ('stationurl', station_data['stationurl'], 'cdata'),
            ('soundenable', 'off'),
            ('sound', -1),
            ('duration', int(duration)),
        ]

        self.get(COMMAND_UIC, 'SetAlarmInfo', params)

    def del_alarm(self, index_list):
        """
        Delete alarm(s).

        Note, speaker's have a limit of 3 alarms and only the first 3 indices will be accepted for deletion while
        remaining will be ignored. This restriction is set on the speaker itself.

        :param index_list: List of alarm indices as returned by get_alarm_info()
        """
        params = [
            ('totaldelnum', len(index_list)),
        ]
        params += [('index', int(i)) for i in index_list]

        self.get(COMMAND_UIC, 'DelAlarm', params)

    def spk_in_group(self, action):
        """
        ???

        :param action: select|?
        """
        params = [('act', action)]

        return self.get(COMMAND_UIC, 'SpkInGroup', params)

    def set_multispk_group(self, name, speakers):
        """
        Group speakers.

        :param name: Group's name
        :param speakers: List of speakers (first one will be treated as main/control one). Dict:
            - name
            - ip
            - mac
        """
        params = [('name', name, 'cdata'), ('index', 1), ('type', 'main'), ('spknum', len(speakers))]

        for i, speaker in enumerate(speakers):
            if i == 0:
                params += [
                    ('audiosourcemacaddr', speaker['mac']),
                    ('audiosourcename', speaker['name'], 'cdata'),
                    ('audiosourcetype', 'speaker'),
                ]
            else:
                params += [
                    ('subspkip', speaker['ip']),
                    ('subspkmacaddr', speaker['mac']),
                ]

        self.get(COMMAND_UIC, 'SetMultispkGroup', params)

    def set_ungroup(self):
        """
        Ungroup speakers.
        """
        self.get(COMMAND_UIC, 'SetUngroup')

    def set_group_name(self, name):
        """
        Update speaker's group name.

        :param name: new name
        """
        params = [('groupname', name, 'cdata')]

        self.get(COMMAND_UIC, 'SetGroupName', params)

    def get_cp_list(self, start_index, list_count):
        """
        Get list of speakers app integrations.

        :returns: List of dicts:
            - cpid - id of app service
            - cpname - service name
            - signinstatus - 0/1
            - username - (optionally if signed in) signed in user name
            - istrial_user - (optionally) 1
        """
        params = [
            ('liststartindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'GetCpList', params)

        return response_list(response['cplist']['cp'])

    def set_cp_service(self, cp_id):
        """
        Switch to a specific cp service.

        It also initiates playback of that service.

        :param cp_id: Cp service id as returned by get_cp_list()
        """
        params = [('cpservice_id', int(cp_id))]

        self.get(COMMAND_CPM, 'SetCpService', params)

    def get_cp_info(self):
        """
        Get info about currently active cp service.

        :returns: Dict
            - cpname
            - timestamp - ISO format
            - category
            - signinstatus - 0/1
            - username
            - subscription_info
            - audioinfo - Dict
                - title
                - streamtype - station|?
                - thumbnail - thumbnail url
                - playstatus - play|pause
            },
        """
        return self.get(COMMAND_CPM, 'GetCpInfo')

    def set_sign_in(self, username, password):
        """
        Authenticate with the currently active service.

        :param username: Service username
        :param password: Service password
        """
        params = [
            ('username', username),
            ('password', password),
        ]

        self.get(COMMAND_CPM, 'SetSignIn', params)

    def set_sign_out(self):
        """
        Sign out from the currently active service.

        You need to be authenticated for this call to be successful.
        """
        self.get(COMMAND_CPM, 'SetSignOut')

    def get_cp_submenu(self):
        """
        Get list of top level service categories.

        :returns: List of dicts:
            - @id - id of the menu item
            - submenuitem_localized - menu item name
        """
        response = self.get(COMMAND_CPM, 'GetCpSubmenu')

        return response_list(response['submenu']['submenuitem'])

    def set_select_cp_submenu(self, content_id, start_index, list_count):
        """
        Get list of sub categories.

        Note, some items are autoplayable. By calling this method you might initiate playback.

        :param content_id: parent id as returned by get_cp_submenu() or this method
        :returns: List of dicts
            - @type
            - title
            - contentid
        """
        params = [
            ('contentid', int(content_id)),
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'SetSelectCpSubmenu', params)

        return response_list(response['menulist']['menuitem'])

    def get_cp_player_playlist(self, start_index, list_count):
        """
        Get currently active service playlist.

        Note, some services have limit of items returned regardless of list_count passed.

        :param start_index:
        :param list_count:
        :returns: List of dicts
            - @type - 1
            - @available - 0/1
            - @currentplaying - (optional) 1 if present
            - artist
            - album
            - mediaid - unique
            - tracklength - track length in seconds
            - title
            - contentid
            - thumbnail - thumbnail url

        """
        params = [
            ('startindex', int(start_index)),
            ('listcount', int(list_count)),
        ]

        response = self.get(COMMAND_CPM, 'GetCpPlayerPlaylist', params)

        return response_list(response['menulist']['menuitem'])

    def set_skip_current_track(self):
        """
        Skip current track and play next item on the playlist.
        """
        self.get(COMMAND_CPM, 'SetSkipCurrentTrack')

    def get_current_play_time(self):
        """
        Get info about current track playback position and length.

        :returns: Dict
            - tracklength - track length in seconds
            - playtime - playback position in seconds
        """
        return self.get(COMMAND_UIC, 'GetCurrentPlayTime')

    def set_play_cp_playlist_track(self, item_id):
        """
        Advance playback to specific track on the playlist.

        :param item_id: Item id as returned by get_cp_player_playlist()
        """
        params = [('selectitemid', int(item_id))]

        return self.get(COMMAND_CPM, 'SetPlayCpPlaylistTrack', params)

    def get_repeat_mode(self):
        """
        Get playback repeat mode.

        :returns: string one|all|off for repeat one track, all tracks on the playlist, or disabled repeat mode
        """
        response = self.get(COMMAND_UIC, 'GetRepeatMode')

        return response['repeat']

    def set_repeat_mode(self, mode):
        """
        Set playback repeat mode.

        :param mode: string one|all|off for repeat one track, all tracks on the playlist, or disabled repeat mode
        """
        params = [('repeatmode', mode)]

        self.get(COMMAND_UIC, 'SetRepeatMode', params)


def on_off_bool(value):
    """Convert on/off to True/False correspondingly."""
    return value == 'on'


def bool_on_off(value):
    """Convert True/False to on/off correspondingly."""
    return 'on' if value else 'off'


def response_list(input_list):
    """xmltodict returns different structure if there's one item on the list."""
    if isinstance(input_list, dict):
        input_list = [input_list]

    return input_list


def format_action(action):
    """Format request action."""
    return '<name>{0}</name>'.format(action)


def format_param(param):
    """
    Format request parameter.

    :param param: Tuple e.g. ('list_count', 30, 'dec')
        - name - name of the param
        - value - mixed value
        - type_hint - (optionally) str|dec|cdata|dec_arr
    """
    (name, value, *attributes) = param

    if not attributes:
        type_hint = 'str'
        if isinstance(value, int):
            type_hint = 'dec'
        if isinstance(value, list) and value and isinstance(value[0], int):
            type_hint = 'dec_arr'
    else:
        type_hint = attributes[0]

    if type_hint == 'cdata':
        return '<p type="{0}" name="{1}" val="empty"><![CDATA[{2}]]></p>'.format(type_hint, name, value)
    if type_hint == 'dec_arr':
        value = ''.join(['<item>{0}</item>'.format(v) for v in value])
        return '<p type="{0}" name="{1}" val="empty">{2}</p>'.format(type_hint, name, value)

    return '<p type="{0}" name="{1}" val="{2}"/>'.format(type_hint, name, value)


def format_payload(action, params=None):
    """Format full request payload."""
    payload = format_action(action)
    if params:
        for param in params:
            payload += format_param(param)

    return payload


def paginator(*args):
    """
    Generator to paginate over api call.

    Api method must accept start_index and list_count parameters.

    :param: callable function to use for pagination
    :param: optionally pass second function that will be used for subsequent pages
    :param: pass all initial values that first callable function accepts, they will be replicated to a second callable
    :returns: Iterable
    """
    if not callable(args[0]):
        raise ValueError('First argument must be a function')

    primary = args[0]
    secondary = args[0]
    args = args[1:]

    if callable(args[0]):
        secondary = args[0]
        args = args[1:]

    import inspect
    primary_parameters = inspect.signature(primary).parameters.keys()
    secondary_parameters = inspect.signature(secondary).parameters.keys()

    # match primary_parameters with args
    primary_kwargs = {}
    secondary_kwargs = {}

    for i, parameter in enumerate(primary_parameters):
        primary_kwargs[parameter] = args[i]
        if parameter in secondary_parameters:
            secondary_kwargs[parameter] = args[i]

    current = primary
    current_kwargs = primary_kwargs
    has_more = True

    while has_more:
        items = current(**current_kwargs)
        for item in items:
            yield item

        has_more = len(items) >= current_kwargs['list_count']

        current = secondary
        current_kwargs = secondary_kwargs
        current_kwargs['start_index'] = current_kwargs['start_index'] + current_kwargs['list_count']



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.equalizer import Equalizer


def get_equalizer():
    api = MagicMock()
    api.get_7band_eq_list.return_value = [
        {
            '@index': '0',
            'presetindex': '0',
            'presetname': 'Pop',
        },
        {
            '@index': '1',
            'presetindex': '1',
            'presetname': 'Rock',
        },
        {
            '@index': '2',
            'presetindex': '2',
            'presetname': 'Custom 1',
        }
    ]
    api.get_current_eq_mode.return_value = {
        'presetindex': '1',
        'presetname': 'Rock',
        'eqvalue1': '4',
        'eqvalue2': '2',
        'eqvalue3': '0',
        'eqvalue4': '0',
        'eqvalue5': '1',
        'eqvalue6': '2',
        'eqvalue7': '1',
    }

    equalizer = Equalizer(api)

    return (equalizer, api)


class TestEqualizer(unittest.TestCase):

    def test_get_presets_names(self):
        equalizer, api = get_equalizer()

        presets_names = equalizer.get_presets_names()

        self.assertEqual(presets_names, ['Pop', 'Rock', 'Custom 1'])

    def test_set_by_preset_name(self):
        equalizer, api = get_equalizer()

        equalizer.set('Rock')

        api.set_7band_eq_mode.assert_called_once_with(1)

    def test_set_by_preset_name_exception(self):
        equalizer, api = get_equalizer()

        self.assertRaises(ValueError, equalizer.set, 'Invalid preset')

        api.set_7band_eq_mode.assert_not_called()

    def test_set_by_bands(self):
        equalizer, api = get_equalizer()

        equalizer.set([1,2,3,4,5,6,-6])

        api.set_7band_eq_value.assert_called_once_with(1, [1,2,3,4,5,6,-6])

    def test_set_by_bands_incorrect(self):
        equalizer, api = get_equalizer()

        self.assertRaises(ValueError, equalizer.set, 1,2,3,4,5,6,6)
        self.assertRaises(ValueError, equalizer.set, [1,2,3])
        self.assertRaises(ValueError, equalizer.set, [1,2,3,4,5,6,6,6,6])
        self.assertRaises(ValueError, equalizer.set, [1,2,3,4,5,6,7])
        self.assertRaises(ValueError, equalizer.set, [1,2,3,4,5,6,-7])
        self.assertRaises(ValueError, equalizer.set, [1,2,3,4,5,6,'Hello there'])

        api.set_7band_eq_value.assert_not_called()

    def test_save_create_new(self):
        equalizer, api = get_equalizer()

        equalizer.save('My new preset')

        api.add_custom_eq_mode.assert_called_once_with(3, 'My new preset')

    def test_save_overwrite_existing(self):
        equalizer, api = get_equalizer()

        equalizer.save('Custom 1')

        api.reset_7band_eq_value.assert_called_once_with(2, [4,2,0,0,1,2,1])

    def test_save_overwrite_current(self):
        equalizer, api = get_equalizer()

        equalizer.save()

        api.reset_7band_eq_value.assert_called_once_with(1, [4,2,0,0,1,2,1])

    def test_delete(self):
        equalizer, api = get_equalizer()

        equalizer.delete('Custom 1')

        api.del_custom_eq_mode.assert_called_once_with(2)

    def test_delete_missing(self):
        equalizer, api = get_equalizer()

        self.assertRaises(ValueError, equalizer.delete, 'Invalid preset')

        api.del_custom_eq_mode.assert_not_called()



import unittest
from unittest.mock import MagicMock

from samsung_multiroom import SamsungSpeakerDiscovery


class TestSpeakerDiscovery(unittest.TestCase):

    @unittest.mock.patch('upnpclient.discover')
    def test_discover(self, upnpclient):
        upnp_devices = [
            MagicMock(),
            MagicMock(),
            MagicMock(),
            MagicMock(),
        ]
        upnpclient.return_value = upnp_devices

        upnp_devices[0].location = 'http://192.168.1.129:7676/smp_3_'
        upnp_devices[1].location = 'http://192.168.1.165:7676/smp_3_'
        upnp_devices[2].location = 'http://192.168.1.216:7676/smp_3_'
        upnp_devices[3].location = 'http://192.168.1.60:7676/smp_7_'

        upnp_devices[0].services = [MagicMock(service_id='urn:samsung.com:serviceId:MultiScreenService')]
        upnp_devices[1].services = [MagicMock(service_id='invalid service')]
        upnp_devices[2].services = [MagicMock(service_id='urn:samsung.com:serviceId:MultiScreenService')]
        upnp_devices[3].services = [MagicMock(service_id='urn:samsung.com:serviceId:MultiScreenService')]

        speaker_discovery = SamsungSpeakerDiscovery()
        speakers = speaker_discovery.discover()

        self.assertEqual(len(speakers), 2)
        self.assertEqual(speakers[0].ip_address, '192.168.1.129')
        self.assertEqual(speakers[1].ip_address, '192.168.1.216')



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import REPEAT_OFF
from samsung_multiroom.service import REPEAT_ONE
from samsung_multiroom.service import NullPlayer


class TestNullPlayer(unittest.TestCase):

    def test_play(self):
        player = NullPlayer()
        self.assertFalse(player.play(MagicMock()))

    def test_jump(self):
        player = NullPlayer()
        player.jump(50)

    def test_resume(self):
        player = NullPlayer()
        player.resume()

    def test_stop(self):
        player = NullPlayer()
        player.stop()

    def test_pause(self):
        player = NullPlayer()
        player.pause()

    def test_next(self):
        player = NullPlayer()
        player.next()

    def test_previous(self):
        player = NullPlayer()
        player.previous()

    def test_get_current_track(self):
        player = NullPlayer()
        track = player.get_current_track()

    def test_repeat(self):
        player = NullPlayer()
        player.repeat(REPEAT_ONE)

    def test_get_repeat(self):
        player = NullPlayer()
        repeat = player.get_repeat()

        self.assertEqual(repeat, REPEAT_OFF)

    def test_is_supported(self):
        player = NullPlayer()

        self.assertTrue(player.is_supported('wifi', 'cp'))
        self.assertTrue(player.is_supported('wifi', 'dlna'))
        self.assertTrue(player.is_supported('bt'))
        self.assertTrue(player.is_supported(None))



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.equalizer import EqualizerGroup


def _get_equalizer_group():
    equalizers = [
        MagicMock(),
        MagicMock(),
        MagicMock(),
    ]

    equalizer_group = EqualizerGroup(equalizers)

    return (equalizer_group, equalizers)


class TestEqualizerGroup(unittest.TestCase):

    def test_get_presets_names_returns_shared(self):
        equalizer_group, equalizers = _get_equalizer_group()

        equalizers[0].get_presets_names.return_value = ['None', 'Pop', 'Jazz', 'Classic', 'Custom 1', 'Custom 2']
        equalizers[1].get_presets_names.return_value = ['None', 'Pop', 'Jazz', 'Classic', 'Custom 2', 'Custom 3']
        equalizers[2].get_presets_names.return_value = ['None', 'Pop', 'Jazz', 'Classic', 'Custom 2', 'Custom 4']

        presets_names = equalizer_group.get_presets_names()

        self.assertEqual(presets_names, ['None', 'Pop', 'Jazz', 'Classic', 'Custom 2'])

    def test_set_sets_all_equalizers(self):
        equalizer_group, equalizers = _get_equalizer_group()

        equalizer_group.set('Jazz')

        equalizers[0].set.assert_called_once_with('Jazz')
        equalizers[1].set.assert_called_once_with('Jazz')
        equalizers[2].set.assert_called_once_with('Jazz')

    def test_save_saves_all_equalizers(self):
        equalizer_group, equalizers = _get_equalizer_group()

        equalizer_group.save('Custom 7')

        equalizers[0].save.assert_called_once_with('Custom 7')
        equalizers[1].save.assert_called_once_with('Custom 7')
        equalizers[2].save.assert_called_once_with('Custom 7')

    def test_delete_deletes_all_equalizers(self):
        equalizer_group, equalizers = _get_equalizer_group()

        equalizer_group.delete('Custom 7')

        equalizers[0].delete.assert_called_once_with('Custom 7')
        equalizers[1].delete.assert_called_once_with('Custom 7')
        equalizers[2].delete.assert_called_once_with('Custom 7')



"""Entry point for service initialisation."""
from ..api import paginator
from .app import AppService
from .dlna import DlnaService
from .tunein import TuneInService


class ServiceRegistry:
    """
    Entry point for service initialisation.
    """

    def __init__(self, api):
        """
        :param api:
        """
        self._api = api

    def get_services_names(self):
        """
        Get list of supported services names.

        :returns: List
        """
        base_names = [
            'dlna',
            'tunein',
        ]

        app_list = self._get_app_list()
        app_names = [s['cpname'] for s in app_list]

        return base_names + app_names

    def service(self, name):
        """
        Get service by name.

        :returns: Service instance
        """
        base_services = {
            'dlna': DlnaService(self._api),
            'tunein': TuneInService(self._api),
        }

        app_list = self._get_app_list()
        app_services = {s['cpname']: AppService(self._api, s['cpid'], s['cpname']) for s in app_list}

        return {**base_services, **app_services}[name]

    def _get_app_list(self):
        return paginator(self._api.get_cp_list, 0, 30)



"""Entry control for speaker operation."""
from .base import SpeakerBase
from .group import SpeakerGroup


class Speaker(SpeakerBase):
    """Entry control for speaker operation."""

    def __init__(self, api, clock, equalizer, player_operator, service_registry):
        """
        Initialise the speaker.

        :param api: SamsungMultiroomApi instance
        :param clock: Clock instance
        :param equalizer: Equalizer instance
        :param player_operator: PlayerOperator instance
        :param service_registry: ServiceRegistry instance
        """
        self._api = api
        self._clock = clock
        self._equalizer = equalizer
        self._player_operator = player_operator
        self._service_registry = service_registry

    @property
    def ip_address(self):
        """
        :returns: Speaker's ip address
        """
        return self._api.ip_address

    @property
    def mac_address(self):
        """
        :returns: Speaker's mac address
        """
        main_info = self._api.get_main_info()
        return main_info['spkmacaddr']

    def get_name(self):
        """
        Retrieve speaker's name.

        :returns: Speaker name string
        """
        return self._api.get_speaker_name()

    def set_name(self, name):
        """
        Set speaker's name.

        :param name: Speaker name string
        """
        self._api.set_speaker_name(name)

    def get_volume(self):
        """
        Get current speaker volume.

        :returns: int current volume
        """
        return self._api.get_volume()

    def set_volume(self, volume):
        """
        Set speaker volume.

        :param volume: speaker volume, integer between 0 and 100
        """
        if not isinstance(volume, int) or int(volume) < 0 or int(volume) > 100:
            raise ValueError('Volume must be integer between 0 and 100')

        self._api.set_volume(volume)

    def get_sources(self):
        """
        Get all supported sources.

        :returns: List of supported sources.
        """
        return ['aux', 'bt', 'hdmi', 'optical', 'soundshare', 'wifi']

    def get_source(self):
        """
        Get currently selected source.

        :returns: selected source string.
        """
        function = self._api.get_func()
        return function['function']

    def set_source(self, source):
        """
        Set speaker source.

        :param source: Speaker source, one of returned by get_sources()
        """
        if source not in self.get_sources():
            raise ValueError('Invalid source {0}'.format(source))

        self._api.set_func(source)

    def is_muted(self):
        """
        Check if speaker is muted.

        :returns: True if muted, False otherwise
        """
        return self._api.get_mute()

    def mute(self):
        """Mute the speaker."""
        self._api.set_mute(True)

    def unmute(self):
        """Unmute the speaker."""
        self._api.set_mute(False)

    def get_services_names(self):
        """
        Get all supported services names.

        :returns: List of strings
        """
        return self._service_registry.get_services_names()

    def service(self, name):
        """
        Get service by type

        :returns: Service instance
        """
        return self._service_registry.service(name)

    @property
    def clock(self):
        """
        Get clock to control time functions.

        :returns: Clock instance
        """
        return self._clock

    @property
    def equalizer(self):
        """
        Get equalizer to control sound adjustments.

        :returns: Equalizer instance
        """
        return self._equalizer

    @property
    def player(self):
        """
        Get currently active player.

        Use player to control playback e.g. pause, resume, get track info etc.

        :returns: Player instance
        """
        return self._player_operator

    def browser(self, name):
        """
        Get media browser by type

        :returns: Browser instance
        """
        return self._service_registry.service(name).browser

    def group(self, name, speakers):
        """
        Group this speaker with another ones.

        This speaker will be the main speaker controlling the playback.

        :param speaker: List of Speaker instances
        :returns: SpeakerGroup instance
        """
        if isinstance(speakers, Speaker):
            speakers = [speakers]

        speakers = [self] + speakers

        speakers_info = []
        for speaker in speakers:
            speakers_info.append({
                'name': speaker.get_name(),
                'ip': speaker.ip_address,
                'mac': speaker.mac_address,
            })

        if not name:
            name = ' + '.join([s['name'] for s in speakers_info])

        self._api.set_multispk_group(name, speakers_info)

        return SpeakerGroup(self._api, name, speakers)

    def ungroup(self):
        """
        Remove this speaker from its current group.
        """
        self._api.set_ungroup()



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.clock import Timer


def get_timer():
    api = MagicMock()

    timer = Timer(api)

    return (timer, api)


class TestTimer(unittest.TestCase):

    def test_start(self):
        timer, api = get_timer()

        timer.start(300)

        api.set_sleep_timer.assert_called_once_with('start', 300)

    def test_stop(self):
        timer, api = get_timer()

        timer.stop()

        api.set_sleep_timer.assert_called_once_with('off', 0)

    def test_get_remaining_time(self):
        timer, api = get_timer()

        api.get_sleep_timer.return_value = {
            'sleepoption': 'start',
            'sleeptime': '270',
        }

        remaining_time = timer.get_remaining_time()

        self.assertEqual(remaining_time, 270)



"""
Init.
"""
from .browser import Browser
from .browser import Item
from .browser import path_to_folders

from .player import Player
from .player import PlayerOperator
from .player import NullPlayer
from .player import Track
from .player import REPEAT_ONE, REPEAT_ALL, REPEAT_OFF

from .service import Service

from .service_registry import ServiceRegistry



import unittest
from unittest.mock import MagicMock

from samsung_multiroom.service import REPEAT_ALL
from samsung_multiroom.service import REPEAT_ONE
from samsung_multiroom.service import NullPlayer
from samsung_multiroom.service import Player
from samsung_multiroom.service import PlayerOperator


def get_mocks():
    api = MagicMock()
    api.get_func.return_value = {'function': 'test_function', 'submode': 'test_submode'}

    player1 = MagicMock(spec=Player, name='player1')
    player1.is_supported.return_value = False
    player2 = MagicMock(spec=Player, name='player2')
    player2.is_supported.return_value = True

    player_operator = PlayerOperator(api, [player1, player2])

    return (player_operator, api, [player1, player2])


class TestPlayerOperator(unittest.TestCase):

    def test_only_player_instances_are_allowed(self):
        api = MagicMock()

        player1 = MagicMock(spec=Player, name='player1')
        player2 = MagicMock(name='player2')

        self.assertRaises(ValueError, PlayerOperator, api, [player1, player2])

    def test_get_player_returns_supporting_player(self):
        api = MagicMock()
        api.get_func.return_value = {'function': 'test_function', 'submode': 'test_submode'}

        player1 = MagicMock(spec=Player, name='player1')
        player1.is_supported.return_value = False
        player2 = MagicMock(spec=Player, name='player2')
        player2.is_supported.return_value = True

        player_operator = PlayerOperator(api, [player1, player2])
        player = player_operator.get_player()

        self.assertEqual(player, player2)
        player1.is_supported.assert_called_once_with('test_function', 'test_submode')
        player2.is_supported.assert_called_once_with('test_function', 'test_submode')

    def test_get_player_returns_nullplayer_if_no_supported(self):
        api = MagicMock()
        api.get_func.return_value = {'function': 'test_function', 'submode': 'test_submode'}

        player1 = MagicMock(spec=Player, name='player1')
        player1.is_supported.return_value = False
        player2 = MagicMock(spec=Player, name='player2')
        player2.is_supported.return_value = False

        player_operator = PlayerOperator(api, [player1, player2])
        player = player_operator.get_player()

        self.assertIsInstance(player, NullPlayer)
        player1.is_supported.assert_called_once_with('test_function', 'test_submode')
        player2.is_supported.assert_called_once_with('test_function', 'test_submode')

    def test_get_play(self):
        api = MagicMock()
        api.get_func.return_value = {'function': 'test_function', 'submode': 'test_submode'}

        player1 = MagicMock(spec=Player, name='player1')
        player1.play.return_value = False
        player2 = MagicMock(spec=Player, name='player2')
        player2.play.return_value = True
        player3 = MagicMock(spec=Player, name='player3')
        player3.play.return_value = True

        playlist = MagicMock()

        player_operator = PlayerOperator(api, [player1, player2, player3])
        self.assertTrue(player_operator.play(playlist))

        player1.play.assert_called_once_with(playlist)
        player2.play.assert_called_once_with(playlist)
        player3.play.assert_not_called()

    def test_get_play_returns_false(self):
        api = MagicMock()
        api.get_func.return_value = {'function': 'test_function', 'submode': 'test_submode'}

        player1 = MagicMock(spec=Player, name='player1')
        player1.play.return_value = False
        player2 = MagicMock(spec=Player, name='player2')
        player2.play.return_value = False
        player3 = MagicMock(spec=Player, name='player3')
        player3.play.return_value = False

        playlist = MagicMock()

        player_operator = PlayerOperator(api, [player1, player2, player3])
        self.assertFalse(player_operator.play(playlist))

        player1.play.assert_called_once_with(playlist)
        player2.play.assert_called_once_with(playlist)
        player3.play.assert_called_once_with(playlist)

    def test_jump(self):
        player_operator, api, players = get_mocks()
        player_operator.jump(50)

        players[0].jump.assert_not_called()
        players[1].jump.assert_called_once_with(50)

    def test_resume(self):
        player_operator, api, players = get_mocks()
        player_operator.resume()

        players[0].resume.assert_not_called()
        players[1].resume.assert_called_once()

    def test_stop(self):
        player_operator, api, players = get_mocks()
        player_operator.stop()

        players[0].stop.assert_not_called()
        players[1].stop.assert_called_once()

    def test_pause(self):
        player_operator, api, players = get_mocks()
        player_operator.pause()

        players[0].pause.assert_not_called()
        players[1].pause.assert_called_once()

    def test_next(self):
        player_operator, api, players = get_mocks()
        player_operator.next()

        players[0].next.assert_not_called()
        players[1].next.assert_called_once()

    def test_previous(self):
        player_operator, api, players = get_mocks()
        player_operator.previous()

        players[0].previous.assert_not_called()
        players[1].previous.assert_called_once()

    def test_repeat(self):
        player_operator, api, players = get_mocks()
        player_operator.repeat(REPEAT_ONE)

        players[0].repeat.assert_not_called()
        players[1].repeat.assert_called_once_with(REPEAT_ONE)

    def test_get_repeat(self):
        player_operator, api, players = get_mocks()

        players[1].get_repeat.return_value = REPEAT_ALL

        repeat = player_operator.get_repeat()

        players[0].get_repeat.assert_not_called()
        players[1].get_repeat.assert_called_once()

    def test_get_current_track(self):
        player_operator, api, players = get_mocks()

        track = MagicMock()
        players[1].get_current_track.return_value = track

        self.assertEqual(player_operator.get_current_track(), track)

        players[0].get_current_track.assert_not_called()
        players[1].get_current_track.assert_called_once()

    def test_is_supported(self):
        player_operator, api, players = get_mocks()
        self.assertTrue(player_operator.is_supported('test_function', 'test_submode'))

        players[0].is_supported.assert_called_once_with('test_function', 'test_submode')
        players[1].is_supported.assert_called_once_with('test_function', 'test_submode')

    def test_is_supported_returns_false(self):
        player_operator, api, players = get_mocks()

        players[0].is_supported.return_value = False
        players[1].is_supported.return_value = False

        self.assertFalse(player_operator.is_supported('test_function', 'test_submode'))

        players[0].is_supported.assert_called_once_with('test_function', 'test_submode')
        players[1].is_supported.assert_called_once_with('test_function', 'test_submode')



"""
Control speaker's sound adjustments.
"""
import abc


class EqualizerBase(metaclass=abc.ABCMeta):
    """
    Abstract base class for equalizers.
    """

    @abc.abstractmethod
    def get_presets_names(self):
        """
        :returns: List of preset names
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def set(self, *args):
        """
        Set equalizer values by preset name or list of band values

        :param: Name of the preset to set or list of 7 band values to set
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def save(self, name=None):
        """
        Create a new or overwrite existing preset.

        :param name: If empty, current preset will be saved, if provided name exist, that preset will overwrite existing
            one, otherwise a new preset is created with that name. In either case band values used will be ones
            currently set on a speaker e.g. via set() method.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def delete(self, name):
        """
        Delete preset.

        :param name: Name of a preset to delete
        """
        raise NotImplementedError()


class Equalizer(EqualizerBase):
    """
    Control speaker's sound adjustments.

    Allows adjustments for bands 150, 300, 600, 1.2k, 2.5k, 5.0k, 10.0k, each value must be integer between -6 and 6.
    """

    def __init__(self, api):
        self._api = api

    def get_presets_names(self):
        """
        :returns: List of preset names
        """
        presets = self._api.get_7band_eq_list()

        return [p['presetname'] for p in presets]

    def set(self, *args):
        """
        Set equalizer values by preset name or list of band values

        :param: Name of the preset to set or list of 7 band values to set
        """
        # set by preset name
        if isinstance(args[0], str):
            preset_index = self._get_preset_index_by_name(args[0])
            if not preset_index:
                raise ValueError('Preset {0} does not exists'.format(args[0]))

            self._api.set_7band_eq_mode(preset_index)
        # set band values
        else:
            if not isinstance(args[0], list) or len(args[0]) != 7:
                raise ValueError('Pass all 7 band values as a list')

            if [value for value in args[0] if not isinstance(value, int) or abs(value) > 6]:
                raise ValueError('Band values must be integers between -6 and 6')

            preset_index = self._get_current_preset()['id']
            self._api.set_7band_eq_value(preset_index, args[0])

    def save(self, name=None):
        """
        Create a new or overwrite existing preset.

        :param name: If empty, current preset will be saved, if provided name exist, that preset will overwrite existing
            one, otherwise a new preset is created with that name. In either case band values used will be ones
            currently set on a speaker e.g. via set() method.
        """
        current_preset = self._get_current_preset()

        # overwrite current one
        if name is None:
            self._api.reset_7band_eq_value(current_preset['id'], current_preset['band_values'])
        else:
            preset_index = self._get_preset_index_by_name(name)

            # overwrite preset with the same name
            if preset_index:
                self._api.reset_7band_eq_value(preset_index, current_preset['band_values'])
            # create a new preset
            else:
                presets = self._api.get_7band_eq_list()
                presets_ids = [int(p['presetindex']) for p in presets]

                self._api.add_custom_eq_mode(max(presets_ids) + 1, name)

    def delete(self, name):
        """
        Delete preset.

        :param name: Name of a preset to delete
        """
        preset_index = self._get_preset_index_by_name(name)

        if not preset_index:
            raise ValueError('Preset {0} does not exists'.format(name))

        self._api.del_custom_eq_mode(preset_index)

    def _get_preset_index_by_name(self, name):
        presets = self._api.get_7band_eq_list()
        matching_presets = [int(p['presetindex']) for p in presets if p['presetname'] == name]

        if matching_presets:
            return matching_presets[0]

        return None

    def _get_current_preset(self):
        preset = self._api.get_current_eq_mode()

        return {
            'id': int(preset['presetindex']),
            'name': preset['presetname'],
            'band_values': [int(preset['eqvalue' + str(i)]) for i in range(1, 8)]
        }



"""Factory for Speaker."""
from .api import SamsungMultiroomApi
from .clock import Alarm
from .clock import Clock
from .clock import Timer
from .equalizer import Equalizer
from .service import PlayerOperator
from .service import ServiceRegistry
from .service.app import AppPlayer
from .service.dlna import DlnaPlayer
from .service.tunein import TuneInPlayer
from .speaker import Speaker


def speaker_factory(ip_address):
    """
    Factory for Speaker.

    :param ip_address: IP address of the speaker.
    """
    api = SamsungMultiroomApi(ip_address)

    timer = Timer(api)
    alarm = Alarm(api)
    clock = Clock(api, timer, alarm)

    equalizer = Equalizer(api)

    players = [
        DlnaPlayer(api),
        TuneInPlayer(api),
        AppPlayer(api),
    ]
    player_operator = PlayerOperator(api, players)

    service_registry = ServiceRegistry(api)

    return Speaker(api, clock, equalizer, player_operator, service_registry)



"""Generic music streaming app service."""
from ..service import Service
from .browser import AppBrowser


class AppService(Service):
    """
    Generic music streaming app service.
    """

    def __init__(self, api, app_id, app_name):
        super().__init__(AppBrowser(api, app_id, app_name))

        self._api = api
        self._id = app_id
        self._name = app_name

    @property
    def name(self):
        """
        :returns: Name of the service
        """
        return self._name

    @property
    def browser(self):
        """
        :returns: Compatible Browser instance for this service
        """
        self._api.set_cp_service(self._id)
        return super().browser

    def login(self, username, password):
        """
        Authenticate with username/password.

        :param username:
        :param password:
        """
        self._api.set_cp_service(self._id)
        self._api.set_sign_in(username, password)

    def logout(self):
        """
        Reset authentication state.
        """
        self._api.set_cp_service(self._id)
        self._api.set_sign_out()
